"""
submit.py

Module for handling submission logic
"""

import requests
import auacm
import argparse
import time
from auacm.utils import subcommand, _find_pid_from_name
from auacm.exceptions import InvalidSubmission

# Map server status results to output strings
RESULTS = {
    'compile': 'Compile Error',
    'runtime': 'Runtime Error',
    'timeout': 'Timeout',
    'wrong': 'Wrong Answer',
    'good': 'Correct Solution'
}

@subcommand('submit')
def submit(args=None, tty=True):
    """Submit a solution"""

    # Some basic argument parsing
    parser = argparse.ArgumentParser(
        add_help=False,
        usage='submit [-p {2,3}] <problem> <file>'
    )
    parser.add_argument('-p', '--python', type=int, choices=[2, 3])
    parser.add_argument('problem')
    parser.add_argument('filename')
    parser.add_argument('-i', '--id', action='store_true')
    args = parser.parse_args(args)


    try:
        submit_file = {'file': open(args.filename, 'rb')}
    except IOError:
        raise InvalidSubmission('Error: Could not open file ' + args.filename)

    if args.id:
        data = {'pid': args.problem}
    else:
        pid = _find_pid_from_name(args.problem)
        if pid == -1:
            raise auacm.exceptions.ProblemNotFoundError(
                'Could not find a problem with the name {}'
                .format(args.problem))
        data = {'pid': pid}

    # Set the python version or default to Python 3
    if args.python:
        data.update({'python': 'py' if args.python == 2 else 'py3'})
    elif args.filename.endswith('.py'):
        # Default to Python 3
        data.update({'python': 'py3'})

    response = requests.post(
        auacm.BASE_URL + 'submit', files=submit_file,
        data=data,
        cookies={'session': auacm.session}
    )

    if not response.ok:
        print('There was an error submitting the solution')
        if 'Unauthorized' in response.text:
            raise auacm.exceptions.UnauthorizedException(
                'You are not logged in')

    return_value = 'Successful submit. Getting results...'
    return_value = 'Running...'
    if tty:
        print(return_value)
    job_id = response.json()['data']['submissionId']
    status = 'start'
    while status == 'start':
        response = requests.get(auacm.BASE_URL + 'submit/' + str(job_id))
        status = response.json()['data']['status']
        if status != 'start':
            return_value += RESULTS[status]
            if tty: print(RESULTS[status])
        else:
            time.sleep(1)

    if not tty: return return_value




"""Utility functions for AUACM package"""

import requests
from auacm.common import BASE_URL

def log(message):
    """Log a message"""
    print(message)

callbacks = dict()

def subcommand(command):
    """Decorator to register a function as a subcommand"""
    def wrapped(function):
        """Add the function to the callbacks"""
        callbacks[command] = function
        return function
    return wrapped


def _find_pid_from_name(name):
    """Look up the pid from the problem name"""
    response = requests.get(BASE_URL + 'problems')
    if not response.ok:
        print('There was an error looking up the problem id')
        exit(1)

    pid = -1
    for problem in response.json()['data']:
        if name.lower() in problem['name'].lower():
            pid = problem['pid']
            break

    return pid



"""
problems.py

Module for handling problem related commands
"""

import requests, argparse, textwrap
import auacm
from auacm.utils import subcommand, _find_pid_from_name

@subcommand('problem')
def problems(args=None):
    """Get all the problems, or search for a specific one"""

    # Some minimal argument parsing
    parser = argparse.ArgumentParser(
        add_help=False,
        usage='problem [-v/--verbose] [-i/--id] <problem>'
    )
    parser.add_argument('-v', '--verbose', action='store_true')
    parser.add_argument('-i', '--id', action='store_true')
    parser.add_argument('problem', nargs='?', default='')
    args = parser.parse_args(args)
    query = args.problem

    # GET request to the API
    request = requests.get(auacm.BASE_URL + 'problems')

    if not request.ok:
        raise auacm.exceptions.ConnectionError(
            'There was an error getting the problems')

    # Filter out problems that aren't similar to the query
    problem_data = request.json()['data']
    results = list()
    for problem in problem_data:
        if not args.id and query.lower() in problem['name'].lower():
            results.append(problem)
        elif args.id and query == str(problem['pid']):
            results.append(problem)

    if not results and query:
        raise auacm.exceptions.ProblemNotFoundError(
            'Could not find problem named {}'.format(query))

    # Print the results
    return_value = ''
    for result in results:
        return_value += result['name'] + '\n'
        if args.verbose:
            return_value += textwrap.dedent("""\
                |    added: {}
                |    appeared: {}
                |    difficulty: {}
                |    pid: {}
                |    shortname: {}
                |    solved: {}
                |    url: {}\n
            """.format(
                result['added'], result['appeared'],
                result['difficulty'], result['pid'],
                result['shortname'], result['solved'], result['url']))

    return return_value.strip()


@subcommand('problem-info')
def get_problem_info(args=None):
    """Get detailed data on a problem (description, input, etc.)"""
    parser = argparse.ArgumentParser(
        add_help=False,
        usage='problem-info [-i/--id] problem'
    )
    parser.add_argument('problem')
    parser.add_argument('-i', '--id', action='store_true')
    args = parser.parse_args(args)

    if args.id:
        pid = args.problem
    else:
        pid = _find_pid_from_name(args.problem)
        if pid == -1:
            raise auacm.exceptions.ProblemNotFoundError(
                'Could not find problem named {}'.format(args.problem))

    response = requests.get(auacm.BASE_URL +  'problems/ ' + str(pid))

    if not response.ok:
        raise auacm.exceptions.ProblemNotFoundError(
            'There was an error getting problem id {}'.format(pid))

    data = response.json()['data']

    # Gather all the results
    return_value = textwrap.dedent('''
        Name: {}

        Description
        {}

        Input
        {}

        Output
        {}
        ''').format(
            data['name'],
            data['description'],
            data['input_desc'],
            data['output_desc']
            )

    for case in data['sample_cases']:
        return_value += textwrap.dedent('''
        Sample Case {}
        Input:
        {}

        Output:
        {}
        ''').format(case['case_num'], case['input'], case['output'])

    return return_value.strip()




"""AUACM top package"""

import auacm.common, auacm.main, auacm.user, auacm.problems, auacm.submit, auacm.competition, auacm.exceptions
from auacm.common import DEBUG, BASE_URL, logo, session
import auacm.utils

# Package metadata
__description__ = 'A command line interface to the Auburn ACM website'
__license__ = 'MIT'
__uri__ = 'https://github.com/BrandonLMorris/auacm-cli'
__version__ = '0.3'
__author__ = 'Brandon Morris'
__email__ = 'brandon.morris95@gmail.com'




"""Custom exception for auacm-cli"""

class ConnectionError(Exception):
    """Represents a failed attempt to connect to the server"""

    def __init__(self, message):
        super(ConnectionError, self).__init__(message)
        self.message = message

    def __str__(self):
        return self.message


class ProblemNotFoundError(Exception):
    """Represents a failed attempt to locate a problem on the server"""

    def __init__(self, message):
        super(ProblemNotFoundError, self).__init__(message)
        self.message = message

    def __str__(self):
        return self.message


class UnauthorizedException(Exception):
    """Represents trying to access a resource they're unauthorized for"""

    def __init__(self, message):
        super(UnauthorizedException, self).__init__(message)
        self.message = message

    def __str__(self):
        return self.message


class InvalidSubmission(Exception):
    """Represents trying to submit an invalid file"""

    def __init__(self, message):
        super(InvalidSubmission, self).__init__(message)
        self.message = message

    def __str__(self):
        return self.message


class CompetitionNotFoundError(Exception):
    """Represents a failed attempt to locate a competition on the server"""

    def __init__(self, message):
        super(CompetitionNotFoundError, self).__init__(message)
        self.message = message

    def __str__(self):
        return self.message




"""
user.py

Module for handling user related commands
"""

import requests, getpass, os
import auacm
from auacm import utils


@utils.subcommand('whoami')
def whoami(args=None):
    """Return basic info about the currently logged in user"""
    r = requests.get(auacm.BASE_URL + 'me', cookies={'session': auacm.session})
    if r.ok:
        return_value = 'User info:\n'
        for key, val in r.json()['data'].items():
            return_value += '    |{} => {}\n'.format(key, val)
        return return_value
    elif r.status_code == 401:
        raise auacm.exceptions.UnauthorizedException('You are not logged in')


@utils.subcommand('logout')
def logout(args=None):
    """Log the user out of the current session"""
    # Erase the contents of the session file
    open(
        os.path.join(os.path.expanduser('~'), '.auacm_session.txt'),
        'w'
    ).close()
    auacm.session = ''


@utils.subcommand('login')
def login(args=None):
    """Log a user in to the website. Keeps up with session data"""
    username = input('Username: ')
    password = getpass.getpass('Password: ')
    response = requests.post(
        auacm.BASE_URL + 'login',
        data={
            'username': username,
            'password': password
        })

    if not response.ok:
        raise auacm.exceptions.ConnectionError(
            'There was an error attempting to log in')

    # Save the new session to a file
    auacm.session = response.cookies['session']
    session_f = open(
        os.path.join(os.path.expanduser('~'), '.auacm_session.txt'),
        'w')
    session_f.write(auacm.session + '\n')
    session_f.close()

    return 'Success!'



"""
main.py

The central entry point of the auacm app.
"""

import requests, sys, textwrap
import auacm
import auacm.utils as utils
from auacm.exceptions import ConnectionError, ProblemNotFoundError, UnauthorizedException, InvalidSubmission, CompetitionNotFoundError

def main(args):
    """
    Entry point for the auacm-cli app

    Supported commands (passed as positional arguments):
        [none]        print logo and help
        test          attempt to connect to the server
        login         log in as a user to the website
        logout        log out of current session
        whoami        print basic info about the current user
        problem       search for a problem
        submit        submit a solution to a problem
        problem-info  get detailed information on a problem
    """
    if not args or args[0] in {'-h', '--help'}:
        # No subcommand, print info
        print(auacm.logo)
        print('Wecome to the Auburn ACM command-line interface!')
        print(textwrap.dedent('''
            Supported Commands:
            [none], -h, --help  print this lovely help
            test                attempt to connect to the server
            login               log into the website
            logout              log out of current session
            whoami              print basic info about the current user
            problem [-v/--verbose] search for a problem
            submit [-i/--id][-p/--python {2, 3}] <problem> <file>
            problem-info [-i/--id] <problem>
            competitions [[-i/--id] <competition>]
        '''))

    elif args[0] in utils.callbacks:
        try:
            print(utils.callbacks[args[0]](args[1:]) or '')
        except (ProblemNotFoundError,
                UnauthorizedException,
                InvalidSubmission,
                CompetitionNotFoundError) as exp:
            print(exp.message)
            exit(1)
        except (requests.exceptions.ConnectionError, ConnectionError):
            print('There was an error connecting to the server: {}'
                  .format(auacm.BASE_URL))
            exit(1)

    else:
        print('Whoops, that subcommand isn\'t supported.\n'
              'Run again with -h or --help to see full list of commands.')


@utils.subcommand('test')
def test(_=None):
    """Try to connect to the server"""
    test_url = auacm.BASE_URL[:-4]
    response = requests.get(test_url)
    if response.ok:
        return 'Connection successful! ' + str(response.status_code)
    else:
        raise auacm.exceptions.ConnectionError('Could not connect to server')


if __name__ == '__main__':
    # Run the app from the main.py script directly
    main(sys.argv[1:])



"""Subcommands related to competitions"""

import auacm, requests, textwrap, argparse
from datetime import datetime
from auacm.utils import subcommand
from auacm.exceptions import CompetitionNotFoundError

@subcommand('competition')
@subcommand('competitions')
def get_comps(args=None):
    """Retrieve one or more competitions from the server"""
    if args:
        return get_one_comp(args)

    response = requests.get(auacm.BASE_URL + 'competitions')

    comps = response.json()['data']
    current = _format_comps(comps['ongoing'])
    upcoming = _format_comps(comps['upcoming'])
    past = _format_comps(comps['past'])

    return textwrap.dedent('''
        [AUACM Competitions]

        Current
        =======
        {}

        Upcoming
        ========
        {}

        Past
        ====
        {}
        ''').format(current, upcoming, past).strip()

def get_one_comp(args):
    """Retrieve info on one specific competition"""
    parser = argparse.ArgumentParser(
        add_help=False,
        usage='competition [-i/--id] <competition>'
    )
    parser.add_argument('-i', '--id', action='store_true')
    parser.add_argument('competition')
    args = parser.parse_args(args)

    if not args.id:
        cid = _cid_from_name(args.competition)
        if cid == -1:
            raise CompetitionNotFoundError(
                'Could not find a competition with the name ' +
                args.competition)
    else:
        cid = args.competition

    response = requests.get(auacm.BASE_URL + 'competitions/' + str(cid))

    if not response.ok or response.status_code == 404:
        raise CompetitionNotFoundError(
            'Could not find competition with id: ' + str(args.competition))

    comp = response.json()['data']

    # 3 Sections: competition, teams, problems
    comp_str = _format_comps([comp['competition']])
    teams = _format_teams(comp['teams'])
    problems = _format_problems(comp['compProblems'])

    return textwrap.dedent('''
        {}

        Teams
        =====
        {}

        Problems
        ========
        {}
        ''').format(comp_str, teams, problems)

def _cid_from_name(comp_name):
    """Return the competition of an id based on it's name"""
    comps = requests.get(auacm.BASE_URL + 'competitions').json()['data']
    for comp in comps['upcoming']:
        if comp_name.lower() in comp['name'].lower():
            return int(comp['cid'])
    for comp in comps['ongoing']:
        if comp_name.lower() in comp['name'].lower():
            return int(comp['cid'])
    for comp in comps['past']:
        if comp_name.lower() in comp['name'].lower():
            return int(comp['cid'])

    return -1

def _format_comps(comps):
    """Return a formatted string for a list of competitions"""
    result = list()
    for comp in comps:
        result.append('{}\t{}\t{}'.format(
            comp['name'], comp['cid'], _get_start_date(comp['startTime'])))

    return '\n'.join(result)

def _format_teams(teams):
    """Return a formatted string of the teams passed in"""
    result = ''
    for team in teams:
        result += team['name'] + '\n'
        if len(team['users']) > 1:
            for user in team['users']:
                result += '|    ' + user + '\n'
        result += '\n'
    return result.strip()

def _format_problems(probs):
    """Return a formatted string of the problems passed in"""
    result = ''
    for label, prob in sorted(probs.items()):
        result += '{}\t{} ({})\n'.format(label, prob['name'], prob['pid'])
    return result.strip()

def _get_start_date(start_time):
    """Return a formatted date string from a competition start time"""
    return datetime.fromtimestamp(start_time).strftime('%m-%d-%Y %H:%M:%S')



import codecs
import os
import re

from setuptools import setup, find_packages

##################################################################

NAME = 'auacm'
PACKAGES = find_packages(where='src')
META_PATH = os.path.join('src', 'auacm', '__init__.py')
KEYWORDS = ['competitive', 'icpc', 'auacm', 'auburn']
CLASSIFIERS = [
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.4'
]
INSTALL_REQUIRES = ['requests']

##################################################################

HERE = os.path.abspath(os.path.dirname(__file__))

def read(*parts):
    """
    Build an absolute file path from *parts* and return the contents
    of the resulting file. Assume UTF-8 encoding
    """
    with codecs.open(os.path.join(HERE, *parts), "rb", "utf-8") as fil:
        return fil.read()

META_FILE = read(META_PATH)

def find_meta(meta):
    """Extract __*meta*__ from META_FILE"""
    meta_match = re.search(
        r"^__{meta}__ = ['\"]([^'\"]*)['\"]".format(meta=meta),
        META_FILE, re.M
    )
    if meta_match:
        return meta_match.group(1)
    raise RuntimeError('Unable to find __{meta}__ string.'.format(meta=meta))

if __name__ == '__main__':
    setup(
        name=NAME,
        description=find_meta('description'),
        license=find_meta('license'),
        url=find_meta('uri'),
        download_url=find_meta('uri') + '/tarball/' + find_meta('version'),
        version=find_meta('version'),
        author=find_meta('author'),
        author_email=find_meta('email'),
        maintainer=find_meta('author'),
        maintainer_email=find_meta('email'),
        keywords=KEYWORDS,
        long_description=read('README.md'),
        packages=PACKAGES,
        package_dir={'': 'src'},
        scripts=['bin/auacm'],
        zip_safe=False,
        classifiers=CLASSIFIERS,
        install_requires=INSTALL_REQUIRES,
    )
